//=============================================================================
// UGoldFix v7 release candidate 3                           Author: Masterkent
//                                                             Date: 2015-08-17
//=============================================================================
class UGoldFix expands Mutator
	config(UGoldFix_v7);

var() const string Version;
var() const int VersionNumber;
var() const int VersionRevision;

const U227ModuleMainClassName = "UGoldFixU227";

// Configurable properties:
// - General:
var(General) config bool bEnableGameFix;
var(General) config bool bEnableMapFix;

// - Advanced [bEnableGameFix]:
var(Advanced_GameFix) config bool bAdjust_bStatic_Actors;
var(Advanced_GameFix) config bool bAdjustBigRocks;
var(Advanced_GameFix) config bool bAdjustBoulders;
var(Advanced_GameFix) config bool bAdjustBrushes;
var(Advanced_GameFix) config bool bAdjustCrushingMovers;
var(Advanced_GameFix) config bool bAdjustDispersionAmmoCharge;
var(Advanced_GameFix) config bool bAdjustDispersionFireRate;
var(Advanced_GameFix) config bool bAdjustDistanceLightnings;
var(Advanced_GameFix) config bool bAdjustEffects;
var(Advanced_GameFix) config bool bAdjustEnhancedSightCheck;
var(Advanced_GameFix) config bool bAdjustGLGrenades;
var(Advanced_GameFix) config bool bAdjustInventoryEffects;
var(Advanced_GameFix) config bool bAdjustIrrelevantMovers;
var(Advanced_GameFix) config bool bAdjustMercenaries;
var(Advanced_GameFix) config bool bAdjustMoverGroups;
var(Advanced_GameFix) config bool bAdjustMoverMovements;
var(Advanced_GameFix) config bool bAdjustPlayersPainTime;
var(Advanced_GameFix) config bool bAdjustSelfDamageMomentum;
var(Advanced_GameFix) config bool bAdjustSpaceMarineCarcasses;
var(Advanced_GameFix) config bool bAdjustSpaceMarineCARDamage;
var(Advanced_GameFix) config bool bAdjustSpaceMarineName;
var(Advanced_GameFix) config bool bAdjustSpaceMarines;
var(Advanced_GameFix) config bool bAdjustSpawnedPlayerState;
var(Advanced_GameFix) config bool bAdjustSpecialEvents;
var(Advanced_GameFix) config bool bAdjustTriggerLights;
var(Advanced_GameFix) config bool bAdjustViewingPlayers;
var(Advanced_GameFix) config bool bAllowAbnormalActors;
var(Advanced_GameFix) config bool bCheckActorPackages;
var(Advanced_GameFix) config bool bDisableMoversGoodCollision;
var(Advanced_GameFix) config bool bExtraPlayerStartLookup;
var(Advanced_GameFix) config bool bInhibitEnvironmentMapping;
var(Advanced_GameFix) config bool bReplaceBlastDecals;
var(Advanced_GameFix) config bool bReplaceExitUPakTeleporters;
var(Advanced_GameFix) config bool bReplaceExplosiveBullets;
var(Advanced_GameFix) config bool bReplaceFlashLightBeams;
var(Advanced_GameFix) config bool bReplaceGrenades;
var(Advanced_GameFix) config bool bReplaceOctagons;
var(Advanced_GameFix) config bool bReplaceRockets;
var(Advanced_GameFix) config bool bReplaceSpinnerProjectiles;
var(Advanced_GameFix) config bool bReplaceUPakRockets;
var(Advanced_GameFix) config bool bReplaceWarlordRockets;

// - Advanced [bEnableMapFix]:
var(Advanced_MapFix) config bool bCoverHolesInLevelsGeometry;
var(Advanced_MapFix) config bool bEnableDecorativeMapChanges;
var(Advanced_MapFix) config bool bEnableLogicMapChanges;
var(Advanced_MapFix) config bool bImproveNPCBehavior;
var(Advanced_MapFix) config bool bModifyNalic2Exit;
var(Advanced_MapFix) config bool bModifyNetRelevance;
var(Advanced_MapFix) config bool bPreventFastRetriggering;
var(Advanced_MapFix) config bool bUnlockNaliSecrets;

// - Multiplayer:
var(Multiplayer) config bool bEnableMoverProtection;
var(Multiplayer) config bool bEnableExtraProtection;

var(Multiplayer) config bool bEnableStationaryEquipment;
var(Multiplayer) config string StationaryEquipment;

var(Multiplayer) config bool bDiscardItemsOnGameEnd_Unreal;
var(Multiplayer) config bool bDiscardItemsOnGameEnd_RTNP;
var(Multiplayer) config string GameEndURL_Unreal;
var(Multiplayer) config string GameEndURL_RTNP;


// Auxiliary properties:
var bool bActiveMutator;
var bool bLevelStartupHandled;
var bool bPlayerPawnsOnly;
var bool bUnrealMap;
var bool bRtnpMap;
var bool bRtnpIntermissionLevel;
var bool bStandardSPMap;
var bool bReplacementFixInitialized;
var name NameTmp;
var SpecialResources LoadedResources;
var UGoldFixSharedInfo SharedInfo;
var UGoldFixU227Base U227Module;

struct StationaryEquipmentTypeInfo
{
	var name PackageName;
	var name ClassName;
};

var StationaryEquipmentTypeInfo StationaryEquipmentClasses[32];

enum E_UEngineVersion
{
	UEngine_vLow,
	UEngine_v224,
	UEngine_v225,
	UEngine_UT,
	UEngine_v226,
	UEngine_v227,
	UEngine_v227g,
	UEngine_v227h,
	UEngine_v227i,
	UEngine_v227j,
	UEngine_vHigh
};

var E_UEngineVersion UEngineVersion;


simulated static function E_UEngineVersion GetUEngineVersion(LevelInfo Level)
{
	local int version;
	local int subversion;

	version = int(Level.EngineVersion);

	if (version == 224)
		return UEngine_v224;
	if (version == 225)
		return UEngine_v225;
	if (version == 226)
		return UEngine_v226;
	if (version == 227)
	{
		subversion = int(Level.GetPropertyText("EngineSubVersion"));

		if (subversion == 7)
			return UEngine_v227g;
		if (subversion == 8)
			return UEngine_v227h;
		if (subversion == 9)
			return UEngine_v227i;
		if (subversion == 10)
			return UEngine_v227j;
		if (subversion > 10)
			return UEngine_vHigh;
		return UEngine_v227;
	}
	if (version >= 400)
		return UEngine_UT;
	if (version > 227)
		return UEngine_vHigh;
	return UEngine_vLow;
}

simulated static function int IntUEngineVersion(E_UEngineVersion x)
{
	return int(x);
}

function InitUEngineVersion()
{
	UEngineVersion = GetUEngineVersion(Level);
}


function PostBeginPlay()
{
	local UGoldFix m;

	foreach AllActors(class'UGoldFix', m)
		if (m.bActiveMutator)
			return;

	bActiveMutator = true;
	ApplyUGoldFix();
}

function ApplyUGoldFix()
{
	InitUEngineVersion();
	ClassifyCurrentMap();

	if (UEngineVersion >= UEngine_v227h)
		LoadU227Module();

	if (bEnableGameFix)
	{
		LoadSpecialResources();
		InitSharedInfo();
		InitReplacementFix();
		InitEffectsFix();

		ApplyServerGameFix();
		ApplyClientGameFix();
	}
	if (bEnableMoverProtection && Level.NetMode != NM_Standalone)
		InitMoverProtection();
	if (bEnableStationaryEquipment && Level.NetMode != NM_Standalone && !Level.Game.bDeathMatch)
		InitStationaryEquipmentList();
	if (bEnableMapFix)
		FixCurrentMap();
	AdjustFinalLevel();

	SaveConfig();

	if (bEnableGameFix)
		SetTimer(0.1, True);
	else
		Disable('Tick');
}

function Timer()
{
	if (!bLevelStartupHandled)
	{
		LevelStartupAdjustments();
		bLevelStartupHandled = True;
	}

	Timer_AdjustActors();
}

function Timer_AdjustActors()
{
	Timer_InitProjectiles();

	Timer_AdjustBigRocks();
	Timer_AdjustMercenaries();
	Timer_AdjustPlayers();
}

function Tick(float DeltaTime)
{
	if (!bEnableGameFix)
		return;
	if (bAdjustDispersionAmmoCharge && !Level.Game.bDeathMatch)
		Tick_AdjustDispersionAmmoCharge();
	if (bAdjustInventoryEffects)
		Tick_AdjustInventoryEffects();
}

function bool CheckReplacement(Actor A, out byte bSuperRelevant)
{
	if (!bActiveMutator || A.bDeleteMe)
		return true;

	if (bEnableGameFix)
	{
		if (bAdjustSpaceMarineName && PlayerReplicationInfo(A) != none && ActorIsA(A.Owner, 'UPak', 'SpaceMarine'))
			PlayerReplicationInfo(A).PlayerName = string(A.Owner.Name);
		if (bAdjustDispersionFireRate &&
			A.class == class'UnrealShare.DispersionAmmo' &&
			!DispersionAmmo(A).bAltFire &&
			A.Instigator != none)
		{
			AdjustDispersionFireRate(A.Instigator.Weapon);
		}
	}
	return AdjustSpawnedActor(A);
}

function AdjustDispersionFireRate(Weapon Weap)
{
	if (Weap == none ||
		Weap.class != class'UnrealShare.DispersionPistol' ||
		Weap.AnimRate >= 4)
	{
		return;
	}
	if (Weap.AnimSequence == 'Shoot1' ||
		Weap.AnimSequence == 'Shoot2' ||
		Weap.AnimSequence == 'Shoot3' ||
		Weap.AnimSequence == 'Shoot4' ||
		Weap.AnimSequence == 'Shoot5')
	{
		Weap.PlayAnim(Weap.AnimSequence, 0.4, 0.2);
	}
}

function bool AdjustSpawnedActor(Actor A)
{
	local string msg;

	if (bPlayerPawnsOnly && Pawn(A) != None && PlayerPawn(A) == None)
		return false;

	if (!bEnableGameFix)
		return true;

	if (Level.NetMode != NM_Standalone &&
		bCheckActorPackages &&
		U227Module != none &&
		bLevelStartupHandled &&
		!U227Module.CheckActorPackage(A))
	{
		if (PlayerPawn(A.Instigator) != none && PlayerPawn(A.Instigator).bAdmin && A.Location != A.Instigator.Location)
		{
			msg = "UGoldFix Warning: New actor of type" @ A.class @ "is irrelevant because package" @
				A.class.outer.name @ "is not included in the ServerPackages map";
			A.Instigator.ClientMessage(msg);
			msg = "    * Note: UGoldFix.bCheckActorPackages == True" $ ", UGoldFix.bAllowAbnormalActors == " $ bAllowAbnormalActors;
			A.Instigator.ClientMessage(msg);
		}
		if (!bAllowAbnormalActors || A.Instigator != none && A.Location == A.Instigator.Location)
			return false;
	}

	if (Effects(A) != none)
		return AdjustEffects(A);
	else if (A.class == class'FlashlightBeam')
		return CreateFlashLightBeamReplacement(FlashLightBeam(A));
	else if (Inventory(A) != none)
	{
		if (bAdjustInventoryEffects)
		{
			if (A.class == class'Amplifier')
				Spawn(class'InventoryEffectsAdjust_Amp').Amp = Amplifier(A);
			else if (A.class == class'ScubaGear' || ActorIsA(A, 'UPak', 'UPakScubaGear'))
				Spawn(class'InventoryEffectsAdjust_Scuba').Scuba = Pickup(A);
		}
	}
	else if (Projectile(A) != none)
	{
		if (LoadedResources != none)
		{
			if (A.class == LoadedResources.class_ExplosiveBullet)
				return ReplaceExplosiveBullet(A);
			if (A.class == LoadedResources.class_GLDetGrenade || A.class == LoadedResources.class_GLGrenade)
			{
				if (bAdjustGLGrenades)
					AdjustClientProjectile(Projectile(A));
				return true;
			}
			if (A.class == LoadedResources.class_SpinnerProjectile)
				return ReplaceSpinnerProjectile(A);
			if (ActorIsA(A, 'UPak', 'UPakRocket'))
				return ReplaceUPakRocket(A);
		}
		if (Boulder1(A) != None)
			AdjustBoulder1(Boulder1(A));
		else if (A.class == class'Grenade')
			return ReplaceGrenade(Grenade(A));
		else if (A.class == class'Rocket' || A.class == class'SeekingRocket')
			return ReplaceRocket(Rocket(A));
		else if (A.class == class'WarlordRocket')
			return ReplaceWarlordRocket(A);
	}
	else if (ActorIsA(A, 'UPak', 'SpaceMarine'))
		AdjustSpaceMarine(Bots(A));
	else if (LoadedResources != none && A.class == LoadedResources.class_SpaceMarineCarcass)
		return CheckSpaceMarineCarcass(CreatureCarcass(A));
	return true;
}

function Class<Weapon> MyDefaultWeapon()
{
	if (bActiveMutator && U227Module == none)
		InitPlayers();
	return Super.MyDefaultWeapon();
}


function name CurrentMap()
{
	return outer.name;
}

function bool ActorIsA(Actor A, name ClassOuterName, name ClassName)
{
	return A != none && A.class.outer.name == ClassOuterName && A.IsA(ClassName);
}

function Actor LoadLevelActor(string ActorName)
{
	return Actor(DynamicLoadObject(Level.outer.name $ "." $ ActorName, class'Actor'));
}

function class<Actor> LoadClass(string ClassName)
{
	return class<Actor>(DynamicLoadObject(ClassName, class'class'));
}


function ClassifyCurrentMap()
{
	local name CurrentMapName;

	bRtnpIntermissionLevel =
		Level.DefaultGameType != none && Level.DefaultGameType.name == 'UPakTransitionInfo';

	if (bRtnpIntermissionLevel)
	{
		bUnrealMap = false;
		bRtnpMap = true;
		return;
	}

	CurrentMapName = CurrentMap();

	if (CurrentMapName == 'Vortex2' ||
		CurrentMapName == 'NyLeve' ||
		CurrentMapName == 'Dig' ||
		CurrentMapName == 'Dug' ||
		CurrentMapName == 'Passage' ||
		CurrentMapName == 'Chizra' ||
		CurrentMapName == 'Ceremony' ||
		CurrentMapName == 'Dark' ||
		CurrentMapName == 'Harobed' ||
		CurrentMapName == 'TerraLift' ||
		CurrentMapName == 'Terraniux' ||
		CurrentMapName == 'Noork' ||
		CurrentMapName == 'Ruins' ||
		CurrentMapName == 'Trench' ||
		CurrentMapName == 'IsvKran4' ||
		CurrentMapName == 'IsvKran32' ||
		CurrentMapName == 'IsvDeck1' ||
		CurrentMapName == 'SpireVillage' ||
		CurrentMapName == 'TheSunspire' ||
		CurrentMapName == 'SkyCaves' ||
		CurrentMapName == 'SkyTown' ||
		CurrentMapName == 'SkyBase' ||
		CurrentMapName == 'VeloraEnd' ||
		CurrentMapName == 'Bluff' ||
		CurrentMapName == 'DasaPass' ||
		CurrentMapName == 'DasaCellars' ||
		CurrentMapName == 'NaliBoat' ||
		CurrentMapName == 'NaliC' ||
		CurrentMapName == 'DCrater' ||
		CurrentMapName == 'ExtremeBeg' ||
		CurrentMapName == 'ExtremeLab' ||
		CurrentMapName == 'ExtremeCore' ||
		CurrentMapName == 'ExtremeGen' ||
		CurrentMapName == 'ExtremeDGen' ||
		CurrentMapName == 'ExtremeDark' ||
		CurrentMapName == 'ExtremeEnd' ||
		CurrentMapName == 'QueenEnd')
	{
		bUnrealMap = true;
		bRtnpMap = false;
		bStandardSPMap = true;
		return;
	}

	if (CurrentMapName == 'DmAriza' ||
		CurrentMapName == 'DmCurse' ||
		CurrentMapName == 'DmDeathFan' ||
		CurrentMapName == 'DmDeck16' ||
		CurrentMapName == 'DmElsinore' ||
		CurrentMapName == 'DmFith' ||
		CurrentMapName == 'DmHealPod' ||
		CurrentMapName == 'DmMorbias' ||
		CurrentMapName == 'DmRadikus' ||
		CurrentMapName == 'DmTundra')
	{
		bUnrealMap = true;
		bRtnpMap = false;
		bStandardSPMap = false;
		return;
	}

	if (CurrentMapName == 'DuskFalls' ||
		CurrentMapName == 'Nevec' ||
		CurrentMapName == 'Eldora' ||
		CurrentMapName == 'Glathriel1' ||
		CurrentMapName == 'Glathriel2' ||
		CurrentMapName == 'Crashsite' ||
		CurrentMapName == 'Crashsite1' ||
		CurrentMapName == 'Crashsite2' ||
		CurrentMapName == 'SpireLand' ||
		CurrentMapName == 'Nagomi' ||
		CurrentMapName == 'Velora' ||
		CurrentMapName == 'NagomiSun' ||
		CurrentMapName == 'Foundry' ||
		CurrentMapName == 'Toxic' ||
		CurrentMapName == 'Glacena' ||
		CurrentMapName == 'Abyss' ||
		CurrentMapName == 'Nalic2')
	{
		bUnrealMap = false;
		bRtnpMap = true;
		bStandardSPMap = true;
		return;
	}

	if (CurrentMapName == 'DmAthena' ||
		CurrentMapName == 'DmDaybreak' ||
		CurrentMapName == 'DmHazard' ||
		CurrentMapName == 'DmStomp' ||
		CurrentMapName == 'DmSunSpeak' ||
		CurrentMapName == 'DmTerra')
	{
		bUnrealMap = false;
		bRtnpMap = true;
		bStandardSPMap = false;
		return;
	}
}


function LoadU227Module()
{
	local class<UGoldFixU227Base> class_UGoldFixU227;

	class_UGoldFixU227 = class<UGoldFixU227Base>(LoadClass(class.outer.name $ "_U227." $ U227ModuleMainClassName));

	if (class_UGoldFixU227 != none)
	{
		U227Module = Spawn(class_UGoldFixU227, self);

		if (U227Module != none)
			U227Module.RegisterServerPackage(string(class.outer.name));
	}
}

function LoadSpecialResources()
{
	LoadedResources = Spawn(class'SpecialResources', self);
}

function InitSharedInfo()
{
	SharedInfo = Spawn(class'UGoldFixSharedInfo', self);
	SharedInfo.LoadedResources = LoadedResources;
	LoadedResources.SharedInfo = SharedInfo;

	SharedInfo.bDisableMoversGoodCollision = bDisableMoversGoodCollision;
	SharedInfo.bInhibitEnvironmentMapping = bInhibitEnvironmentMapping;
	SharedInfo.bReplaceBlastDecals = bReplaceBlastDecals;
}

function InitReplacementFix()
{
	if (bAdjustBoulders)
		MakeNonGameRelevant("UnrealI.Boulder1");

	if (bAdjustDispersionFireRate)
		MakeNonGameRelevant("UnrealShare.DispersionAmmo");

	if (bReplaceExplosiveBullets && LoadedResources.class_ExplosiveBullet != none)
		MakeNonGameRelevant("UPak.ExplosiveBullet");

	if (bAdjustGLGrenades &&
		LoadedResources.class_GLDetGrenade != none &&
		LoadedResources.class_GLGrenade != none)
	{
		MakeNonGameRelevant("UPak.GLDetGrenade");
		MakeNonGameRelevant("UPak.GLGrenade");
	}

	if (bReplaceGrenades)
		MakeNonGameRelevant("UnrealShare.Grenade");

	if (bReplaceOctagons && LoadedResources.class_Octagon != none && Level.NetMode != NM_Standalone)
		MakeNonGameRelevant("UPak.Octagon");

	if (bReplaceRockets)
	{
		MakeNonGameRelevant("UnrealShare.Rocket");
		MakeNonGameRelevant("UnrealShare.SeekingRocket");
	}

	if (bReplaceSpinnerProjectiles && UEngineVersion <= UEngine_v227i &&
		LoadedResources.class_SpinnerProjectile != none && Level.NetMode != NM_Standalone)
	{
		MakeNonGameRelevant("UPak.SpinnerProjectile");
	}

	if (bReplaceUPakRockets &&
		LoadedResources.class_UPakRocket != none &&
		LoadedResources.class_TowRocket != none)
	{
		MakeNonGameRelevant("UPak.TowRocket");
		MakeNonGameRelevant("UPak.UPakRocket");
	}

	if (bReplaceWarlordRockets)
		MakeNonGameRelevant("UnrealI.WarlordRocket");

	bReplacementFixInitialized = true;
}

function InitEffectsFix()
{
	if (UEngineVersion == UEngine_v226 && bAdjustEffects && Level.NetMode == NM_DedicatedServer)
	{
		MakeNonGameRelevant("UnrealShare.RingExplosion");
		MakeNonGameRelevant("UnrealShare.RingExplosion2");
		MakeNonGameRelevant("UnrealShare.RingExplosion3");
		MakeNonGameRelevant("UnrealShare.RingExplosion4");
		MakeNonGameRelevant("UnrealShare.SmallSpark");
		MakeNonGameRelevant("UnrealShare.SmallSpark2");
		MakeNonGameRelevant("UnrealShare.Spark3");
		MakeNonGameRelevant("UnrealShare.Spark32");
		MakeNonGameRelevant("UnrealShare.Spark33");
		MakeNonGameRelevant("UnrealShare.Spark34");
		MakeNonGameRelevant("UnrealShare.Spark35");
		MakeNonGameRelevant("UnrealShare.WaterRing");
	}
}

function ApplyServerGameFix()
{
	if (bAdjust_bStatic_Actors)
		Fix_bStatic_Actors();
	if (bAdjustBrushes)
		FixBrushes();
	if (bAdjustCrushingMovers)
		AdjustCrushingMovers();
	if (bAdjustDistanceLightnings)
		FixDistanceLightnings();
	if (bAdjustEnhancedSightCheck)
		AdjustEnhancedSightCheck();
	if (bReplaceExitUPakTeleporters)
		FixExitTeleporters();
	if (bAdjustSpecialEvents)
		FixSpecialEvents();
	if (bAdjustMoverGroups)
		AdjustMoverGroups();
	if (bAdjustMoverMovements)
		AdjustMoverMovements();
	if (bDisableMoversGoodCollision)
		DisableMoversGoodCollision(Level);
	if (UEngineVersion >= UEngine_v227h)
		ApplyU227Fix();
}

function ApplyClientGameFix()
{
	Spawn(class'UGoldClientFix', self).SharedInfo = SharedInfo;
}

function Fix_bStatic_Actors()
{
	local Actor A;
	local Trigger Replacement;
	local int i;

	if (Level.NetMode == NM_Standalone)
		return;

	foreach AllActors(Class'Actor',A)
		if (!A.bStatic && A.default.bStatic && !A.bNoDelete && A.RemoteRole != ROLE_None && !A.bHidden)
		{
			Replacement = Spawn(class'Trigger',,, A.Location, A.Rotation);
			if (Replacement == none)
			{
				A.Destroy();
				continue;
			}
			// Advanced
			Replacement.bAlwaysRelevant = A.bAlwaysRelevant;
			Replacement.bCanTeleport = A.bCanTeleport;
			Replacement.bHidden = A.bHidden;
			Replacement.LifeSpan = A.LifeSpan;

			// Collision
			Replacement.SetCollisionSize(A.CollisionRadius, A.CollisionHeight);
			Replacement.SetCollision(A.bCollideActors, A.bBlockActors, A.bBlockPlayers);
			Replacement.bCollideWorld = A.bCollideWorld;

			// Display
			Replacement.bMeshEnviromap = A.bMeshEnviromap;
			Replacement.bUnlit = A.bUnlit;
			Replacement.DrawScale = A.DrawScale;
			Replacement.DrawType = A.DrawType;
			Replacement.Fatness = A.Fatness;
			Replacement.Mesh = A.Mesh;
			for (i = 0; i < ArrayCount(A.Multiskins); ++i)
				Replacement.Multiskins[i] = A.Multiskins[i];
			Replacement.ScaleGlow = A.ScaleGlow;
			Replacement.Skin = A.Skin;
			Replacement.Sprite = A.Sprite;
			Replacement.Style = A.Style;
			Replacement.Texture = A.Texture;

			// Events
			Replacement.Tag = A.Tag;

			// LightColor
			Replacement.LightBrightness = A.LightBrightness;
			Replacement.LightHue = A.LightHue;
			Replacement.LightSaturation = A.LightSaturation;

			// Lighting
			Replacement.bActorShadows = A.bActorShadows;
			Replacement.bCorona = A.bCorona;
			Replacement.bLensFlare = A.bLensFlare;
			Replacement.bSpecialLit = A.bSpecialLit;
			Replacement.LightCone = A.LightCone;
			Replacement.LightEffect = A.LightEffect;
			Replacement.LightPeriod = A.LightPeriod;
			Replacement.LightPhase = A.LightPhase;
			Replacement.LightRadius = A.LightRadius;
			Replacement.LightType = A.LightType;
			Replacement.VolumeBrightness = A.VolumeBrightness;
			Replacement.VolumeFog = A.VolumeFog;
			Replacement.VolumeRadius = A.VolumeRadius;

			// Movement
			Replacement.SetBase(A.Base);
			Replacement.SetPhysics(A.Physics);
			Replacement.AttachTag = A.AttachTag;
			Replacement.bBounce = A.bBounce;
			Replacement.bFixedRotationDir = A.bFixedRotationDir;
			Replacement.Buoyancy = A.Buoyancy;
			Replacement.Velocity = A.Velocity;

			// Networking
			Replacement.RemoteRole = ROLE_SimulatedProxy;

			// Sound
			Replacement.AmbientSound = A.AmbientSound;
			Replacement.SoundPitch = A.SoundPitch;
			Replacement.SoundRadius = A.SoundRadius;
			Replacement.SoundVolume = A.SoundVolume;
			Replacement.TransientSoundRadius = A.TransientSoundRadius;
			Replacement.TransientSoundVolume = A.TransientSoundVolume;

			A.Destroy();
		}
}

function FixBrushes()
{
	local Brush br;

	if (Level.NetMode == NM_Standalone)
		return;

	foreach AllActors(class'Brush', br)
		if (!br.bNoDelete && !br.bStatic)
			br.Destroy();
}

function AdjustCrushingMovers()
{
	local Mover m;

	foreach AllActors(class'Mover', m)
		if (m.MoverEncroachType == ME_CrushWhenEncroach)
		{
			m.MoverEncroachType = ME_IgnoreWhenEncroach;
			m.EncroachDamage = Max(m.EncroachDamage, 100000000);
		}
}

function FixDistanceLightnings()
{
	local DistanceLightning dl;

	if (Level.NetMode == NM_Standalone)
		return;
	foreach AllActors(class'Engine.DistanceLightning', dl)
	{
		dl.bAlwaysRelevant = true;
		if (UEngineVersion >= UEngine_v227)
		{
			dl.RemoteRole = ROLE_DumbProxy;
			if (Level.NetMode == NM_DedicatedServer)
				dl.SetTimer(5 + FRand() * 10, false);
		}
	}
}

function AdjustEnhancedSightCheck()
{
	// AI fix (Issue: creatures can see through non-transparent movers or
	// behave weird in presence of several players)
	if (bAdjustEnhancedSightCheck &&
		Level.NetMode != NM_Standalone &&
		UEngineVersion >= UEngine_v227i)
	{
		Level.Game.SetPropertyText("bAlwaysEnhancedSightCheck", "false");
	}
}

function FixExitTeleporters()
{
	local Teleporter OldTelep;
	local NonStaticTeleporter NewTelep;

	if (Level.NetMode == NM_Standalone || UEngineVersion >= UEngine_v227g)
		return;

	foreach AllActors(class'Teleporter', OldTelep)
	{
		if (ActorIsA(OldTelep, 'UPak', 'UPakTeleporter'))
		{
			NewTelep = Spawn(class'NonStaticTeleporter',,, OldTelep.Location, OldTelep.Rotation);
			NewTelep.URL = OldTelep.URL;
			NewTelep.bEnabled = OldTelep.bEnabled;
			NewTelep.Tag = OldTelep.Tag;
			NewTelep.SetCollision(OldTelep.bCollideActors, OldTelep.bBlockActors, OldTelep.bBlockPlayers);
			NewTelep.SetCollisionSize(OldTelep.CollisionRadius, OldTelep.CollisionHeight);
			NewTelep.PrototypeTeleporter = OldTelep;
			OldTelep.SetCollision(false, false, false);
		}
	}
}

function FixSpecialEvents()
{
	local SpecialEvent spec_event;

	if (Level.NetMode != NM_Standalone)
		foreach AllActors(class'SpecialEvent', spec_event)
		{
			if (spec_event.TransientSoundVolume > 12000)
				spec_event.TransientSoundVolume = 12000;
		}
}

function AdjustMoverGroups()
{
	local Mover m;

	if (UEngineVersion > UEngine_v227i)
		return;

	foreach AllActors(class'Mover', m)
	{
		if (m.ReturnGroup == '' && (m.Tag == m.class.name || m.Tag == '') ||
			m.ReturnGroup == m.class.name)
		{
			m.Leader = m;
			m.Follower = none;
		}
	}
}

function AdjustMoverMovements()
{
	local Mover m;

	if (Level.NetMode == NM_Standalone)
		return;

	foreach AllActors(class'Mover', m)
		if (Level.Game.IsRelevant(m))
		{
			m.bAlwaysRelevant = true;
			Spawn(class'MoverMovementSynchronizer').ControlledMover = m;
		}
}

static function DisableMoversGoodCollision(LevelInfo Level)
{
	local Mover m;

	if (class'MapFixBase'.static.MoverGoodCollisionSupported(Level))
	{
		foreach Level.AllActors(class'Mover', m)
			m.SetPropertyText("bUseGoodCollision", "false");
	}
}

function ApplyU227Fix()
{
	if (U227Module != none)
		U227Module.AddGameRules();
}


// Prevents Movers from being locked by other actors
// Rationale: locking some key Movers could make completing level impossible
function InitMoverProtection()
{
	local Mover m;

	if (!AddMoverProtectionMutator())
	{
		foreach AllActors(class'Mover', m)
			if (m.MoverEncroachType == ME_ReturnWhenEncroach || m.MoverEncroachType == ME_StopWhenEncroach)
				m.MoverEncroachType = ME_IgnoreWhenEncroach;
	}
}

function bool AddMoverProtectionMutator()
{
	local class<Mutator> MoverProtectionMutatorClass;
	MoverProtectionMutatorClass =
		class<Mutator>(LoadClass("MoverProtection.MoverProtection"));
	if (MoverProtectionMutatorClass != None)
	{
		Level.Game.BaseMutator.AddMutator(Spawn(MoverProtectionMutatorClass));
		return true;
	}
	return false;
}

function InitStationaryEquipmentList()
{
	local int i, n;
	local int space_pos, dot_pos;
	local string EquipmentList;
	local string ClassName;

	if (StationaryEquipment == "")
		return;

	EquipmentList = StationaryEquipment;
	i = 0;
	n = ArrayCount(StationaryEquipmentClasses);

	while (i < n)
	{
		space_pos = InStr(EquipmentList, " ");
		if (space_pos < 0)
			ClassName = EquipmentList;
		else
		{
			ClassName = Left(EquipmentList, space_pos);
			EquipmentList = Mid(EquipmentList, space_pos + 1);
		}
		dot_pos = InStr(ClassName, ".");
		if (dot_pos < 0)
		{
			StationaryEquipmentClasses[i].PackageName = '';
			StationaryEquipmentClasses[i].ClassName = ConvertStringToName(ClassName);
		}
		else
		{
			StationaryEquipmentClasses[i].PackageName = ConvertStringToName(Left(ClassName, dot_pos));
			StationaryEquipmentClasses[i].ClassName = ConvertStringToName(Mid(ClassName, dot_pos + 1));
		}

		if (StationaryEquipmentClasses[i].ClassName != '')
			++i;

		if (space_pos < 0)
			break;
	}
}


function FixCurrentMap()
{
	if (bUnrealMap)
		Spawn(class'MapFixForUnreal', self);
	else if (bRtnpMap)
		Spawn(class'MapFixForRTNP', self);
}

function AdjustFinalLevel()
{
	local string GameEndURL;
	local Teleporter Telep, ExitTelep;

	if (Level.NetMode == NM_Standalone)
		return;

	if (CurrentMap() == 'QueenEnd')
	{
		ExitTelep = Teleporter(LoadLevelActor("Teleporter0"));
		GameEndURL = GameEndURL_Unreal;
		if (bDiscardItemsOnGameEnd_Unreal)
			Spawn(class'DiscardItemsOnLevelEnd');
	}
	else if (CurrentMap() == 'Nalic2')
	{
		ExitTelep = Teleporter(LoadLevelActor("UPakTeleporter0"));
		GameEndURL = GameEndURL_RTNP;
		if (bDiscardItemsOnGameEnd_RTNP)
			Spawn(class'DiscardItemsOnLevelEnd');
	}
	else
		return;

	if (GameEndURL == "")
		return;

	foreach AllActors(class'Teleporter', Telep)
		if (Telep == ExitTelep || NonStaticTeleporter(Telep) != none && NonStaticTeleporter(Telep).PrototypeTeleporter == ExitTelep)
		{
			Telep.URL = GameEndURL;
		}
}

function LevelStartupAdjustments()
{
	if (bAdjustIrrelevantMovers)
		FixIrrelevantMovers();
	if (bAdjustTriggerLights)
		FixTriggerLights();
	if (bCheckActorPackages && !bAllowAbnormalActors)
		CheckActorPackages();
	if (bEnableStationaryEquipment && Level.NetMode != NM_Standalone && !Level.Game.bDeathMatch)
		WrapStationaryEquipment();
}

function FixIrrelevantMovers()
{
	local Mover m;

	foreach AllActors(class'Engine.Mover', m)
		if (!Level.Game.IsRelevant(m))
		{
			if (m.Brush != none)
			{
				m.bHidden = False;
				m.DrawType = DT_Brush;
			}
			else
				m.bHidden = true;
			m.SetLocation(vect(0, 0, -100000.0));
			m.Tag = '';
			m.RemoteRole = ROLE_DumbProxy;
		}
}

function FixTriggerLights()
{
	local TriggerLight tl;

	if (Level.NetMode != NM_Standalone)
	{
		foreach AllActors(class'Engine.TriggerLight', tl)
			if (tl.Region.Zone != none && tl.Region.Zone.IsA('SkyZoneInfo'))
				tl.bAlwaysRelevant = true;
	}
}

function CheckActorPackages()
{
	local Inventory anInventory;
	local Pawn aPawn;

	if (Level.NetMode == NM_Standalone || U227Module == none)
		return;

	foreach AllActors(class'Inventory', anInventory)
		if (!U227Module.CheckActorPackage(anInventory))
			anInventory.Destroy();

	foreach AllActors(class'Pawn', aPawn)
		if (!U227Module.CheckActorPackage(aPawn))
			aPawn.Destroy();
}

function WrapStationaryEquipment()
{
	local Inventory inv;

	foreach AllActors(class'Inventory', inv)
		if (inv.Owner == none && !inv.bHeldItem && IsStationaryEquipment(inv))
			Spawn(class'InventoryTrigger').AttachInventory(inv);
}

function MakeNonGameRelevant(string ClassName)
{
	LoadClass(ClassName);
	ConsoleCommand("set " $ ClassName $ " bGameRelevant false");
}

function Timer_InitProjectiles()
{
	if (!bReplacementFixInitialized)
		InitReplacementFix(); // will be called for loaded games
}

function Timer_AdjustBigRocks()
{
	local BigRock rock;
	local float AdjustedCollisionHeight;

	if (!bAdjustBigRocks)
		return;

	foreach AllActors(class'BigRock', rock)
	{
		if (Boulder1(rock) == None && (rock.Physics == PHYS_None || VSize(rock.velocity) < 60 - rock.DrawScale))
		{
			AdjustedCollisionHeight =
				0.5 * rock.default.CollisionHeight * rock.DrawScale / rock.default.DrawScale;

			if (rock.Physics == PHYS_None)
			{
				if (rock.CollisionHeight <= AdjustedCollisionHeight + 0.1)
					return;
				rock.SetPhysics(PHYS_Falling);
			}
			
			rock.SetCollisionSize(
				0.5 * rock.default.CollisionRadius * rock.DrawScale / rock.default.DrawScale,
				AdjustedCollisionHeight);
		}
	}
}

function Timer_AdjustMercenaries()
{
	local Mercenary merc;

	if (!bAdjustMercenaries || UEngineVersion > UEngine_v227i)
		return;

	foreach AllActors(class'Mercenary', merc)
		if (!merc.bIsInvulnerable &&
			merc.PainTime <= 0 &&
			(merc.Region.Zone.bPainZone || merc.FootRegion.Zone.bPainZone || merc.HeadRegion.Zone.bPainZone))
		{
			merc.PainTime = 0.01;
		}
}

function Timer_AdjustPlayers()
{
	local PlayerPawn player;

	foreach AllActors(class'PlayerPawn', player)
	{
		if (bAdjustPlayersPainTime && player.PainTime > 2 && player.FootRegion.Zone.bPainZone)
			player.PainTime = 0.5;
		if (Level.NetMode != NM_Standalone && !Level.Game.bDeathMatch &&
			bRtnpMap && bStandardSPMap &&
			UEngineVersion <= UEngine_v226)
		{
			CheckScubaGear(player);
		}
	}
}


function Tick_AdjustDispersionAmmoCharge()
{
	local PlayerPawn Player;

	foreach AllActors(class'PlayerPawn', Player)
		if (Player.Weapon != none &&
			Player.Weapon.class == class'UnrealShare.DispersionPistol' &&
			Player.Weapon.AmmoType != none &&
			Player.Weapon.AmmoType.AmmoAmount < 1 &&
			!Player.Weapon.IsInState('AltFiring'))
		{
			Player.Weapon.AmmoType.AmmoAmount = 1;
		}
}

function Tick_AdjustInventoryEffects()
{
	local PlayerPawn Player;

	foreach AllActors(class'PlayerPawn', Player)
	{
		if (Player.Weapon != none && Player.Handedness != 2)
			Player.Weapon.bHideWeapon = Player.bHidden;
	}
}


function bool IsStationaryEquipment(Inventory Inv)
{
	local int i, n;
	n = ArrayCount(StationaryEquipmentClasses);
	for (i = 0; i < n && StationaryEquipmentClasses[i].ClassName != ''; ++i)
		if (Inv.class.name == StationaryEquipmentClasses[i].ClassName)
		{
			if (StationaryEquipmentClasses[i].PackageName != '')
			{
				if (Inv.class.outer.name == StationaryEquipmentClasses[i].PackageName)
					return true;
			}
			else if (Inv.class.outer.name == 'UnrealI' ||
				Inv.class.outer.name == 'UnrealShare' ||
				Inv.class.outer.name == 'UPak')
			{
				return true;
			}
		}
	return false;
}

function AdjustClientProjectile(Projectile Proj)
{
	if (Level.NetMode != NM_Standalone)
		Spawn(class'ProjectileClientAdjustment').ControlledProjectile = Proj;
}

function bool AdjustEffects(Actor A)
{
	if (LoadedResources != none && A.class == LoadedResources.class_Octagon && LoadedResources.class_Octagon != none)
		return ReplaceOctagon(A);

	if (bAdjustEffects && UEngineVersion == UEngine_v226 && Level.NetMode == NM_DedicatedServer &&
		A.bNetTemporary && A.RemoteRole == ROLE_SimulatedProxy)
	{
		A.RemoteRole = ROLE_Authority;
	}

	return true;
}

function bool ReplaceOctagon(Actor A)
{
	local UGoldFixSharedInfoContainer shared_info_container;

	if (!bReplaceOctagons || Level.NetMode == NM_Standalone)
		return true;

	shared_info_container = Spawn(class'UGoldFixSharedInfoContainer', A.Owner);
	shared_info_container.SharedInfo = SharedInfo;
	A.Spawn(class'UPak_Octagon', shared_info_container,, A.Location, A.Rotation);

	A.SetLocation(vect(0, 0, -100000));
	return false;
}

function bool CreateFlashLightBeamReplacement(FlashLightBeam A)
{
	if (bReplaceFlashLightBeams && (Level.NetMode != NM_Standalone || UEngineVersion <= UEngine_v227i))
		class'UnrealShare_FlashLightBeam'.static.CreateReplacement(A);
	return true;
}

function AdjustBoulder1(Boulder1 rock)
{
	if (!bAdjustBoulders)
		return;
	rock.Damage = 6850;
	rock.SetPhysics(PHYS_Projectile);
}

function bool ReplaceExplosiveBullet(Actor A)
{
	local UGoldFixSharedInfoContainer shared_info_container;
	local UPak_ExplosiveBullet proj;

	if (!bReplaceExplosiveBullets)
		return true;

	shared_info_container = Spawn(class'UGoldFixSharedInfoContainer', A.Owner);
	shared_info_container.SharedInfo = SharedInfo;
	proj = A.Spawn(class'UPak_ExplosiveBullet', shared_info_container,, A.Location, A.Rotation);
	if (proj != none)
		proj.bAdjustSpaceMarineCARDamage = bAdjustSpaceMarineCARDamage;

	A.SetLocation(vect(0, 0, -100000));
	return false;
}

function bool ReplaceGrenade(Grenade OriginalGrenade)
{
	local UnrealShare_Grenade_Replacer replacer;

	if (!bReplaceGrenades)
		return true;

	replacer = OriginalGrenade.Spawn(class'UnrealShare_Grenade_Replacer');
	replacer.OriginalGrenade = OriginalGrenade;
	replacer.SharedInfo = SharedInfo;
	replacer.ReplaceGrenade();

	return true;
}

function bool ReplaceRocket(Rocket OriginalRocket)
{
	local UnrealShare_Rocket_Replacer replacer;

	if (!bReplaceRockets)
		return true;

	replacer = OriginalRocket.Spawn(class'UnrealShare_Rocket_Replacer');
	replacer.OriginalRocket = OriginalRocket;
	replacer.SharedInfo = SharedInfo;
	replacer.ReplaceRocket();

	return true;
}

function bool ReplaceSpinnerProjectile(Actor A)
{
	local UGoldFixSharedInfoContainer shared_info_container;

	if (!bReplaceSpinnerProjectiles || Level.NetMode == NM_Standalone || UEngineVersion > UEngine_v227i)
		return true;

	shared_info_container = Spawn(class'UGoldFixSharedInfoContainer', A.Owner);
	shared_info_container.SharedInfo = SharedInfo;
	A.Spawn(class'UPak_SpinnerProjectile', shared_info_container,, A.Location, A.Rotation);

	A.SetLocation(vect(0, 0, -100000));
	return false;
}

function bool ReplaceUPakRocket(Actor A)
{
	local UGoldFixSharedInfoContainer shared_info_container;
	local class<Actor> rocket_class;

	if (!bReplaceUPakRockets)
		return true;

	if (A.class.name == 'TowRocket')
		rocket_class = class'UPak_TowRocket';
	else
		rocket_class = class'UPak_UPakRocket';

	shared_info_container = Spawn(class'UGoldFixSharedInfoContainer', A.Owner);
	shared_info_container.SharedInfo = SharedInfo;
	A.Spawn(rocket_class, shared_info_container,, A.Location, A.Rotation);

	A.SetLocation(vect(0, 0, -100000));
	return false;
}

function bool ReplaceWarlordRocket(Actor A)
{
	local UnrealI_WarlordRocket new_rocket;

	if (!bReplaceWarlordRockets)
		return true;

	new_rocket = A.Spawn(class'UnrealI_WarlordRocket', A.Owner,, A.Location, A.Rotation);
	new_rocket.SharedInfo = SharedInfo;

	A.SetLocation(vect(0, 0, -100000));
	return false;
}

function AdjustSpaceMarine(Bots creature)
{
	local SpaceMarineController controller;

	if (!bAdjustSpaceMarines || creature == none)
		return;
	controller = Spawn(class'SpaceMarineController');
	controller.ControlledPawn = creature;
	controller.Fix = self;
}

function bool CheckSpaceMarineCarcass(CreatureCarcass A)
{
	local Carcass c;

	if (UEngineVersion > UEngine_v227i)
		return true;

	if (bAdjustSpaceMarineCarcasses)
	{
		foreach A.AllActors(class'Engine.Carcass', c)
			if (c != A && c.Instigator == A.instigator)
				return false;
	}
	if (ActorIsA(A.Instigator, 'UPak', 'SpaceMarine'))
		A.Instigator.bIsPlayer = true;
	return true;
}


function InitPlayers()
{
	local PlayerPawn P;
	foreach AllActors(class'PlayerPawn', P)
		InitPlayer(P);
}

function InitPlayer(Pawn P)
{
	if (Level.NetMode != NM_Standalone && !Level.Game.bDeathMatch)
		GiveInitialInventoryTo(P);
	AdjustViewingPlayersFor(P);
}

function GiveInitialInventoryTo(Pawn Player)
{
	if (IsSpectator(Player) || Player.Health <= 0)
		return;
	if (bRtnpMap && bStandardSPMap)
		CheckScubaGear(Player);
}

function bool IsSpectator(Pawn Player)
{
	return Player.PlayerReplicationInfo != none && Player.PlayerReplicationInfo.bIsSpectator;
}

// pre-227 versions don't support StringToName
function name ConvertStringToName(string S)
{
	SetPropertyText("NameTmp", S);
	return NameTmp;
}

function CheckScubaGear(Pawn Player)
{
	local Pickup newScubaGear;

	if (IsSpectator(Player) || Player.Health <= 0 ||
		LoadedResources == none ||
		LoadedResources.class_UPakScubaGear == none ||
		Player.FindInventoryType(LoadedResources.class_UPakScubaGear) != none)
	{
		return;
	}
	
	newScubaGear = Spawn(LoadedResources.class_UPakScubaGear,,, Location);

	if (newScubaGear != none)
	{
		newScubaGear.bHeldItem = true;
		newScubaGear.GiveTo(Player);
		Player.SelectedItem = newScubaGear;
		newScubaGear.PickupFunction(Player);
	}
}

function AdjustViewingPlayersFor(Pawn Player)
{
	if (bEnableGameFix &&
		bAdjustViewingPlayers &&
		Level.NetMode != NM_Standalone &&
		!Level.Game.bTeamGame &&
		Player.PlayerReplicationInfo != none)
	{
		Player.PlayerReplicationInfo.TeamID = -1;
	}
}

static function Projectile_HurtRadius(
	UGoldFixSharedInfo SharedInfo,
	Projectile Proj,
	float DamageAmount,
	float DamageRadius,
	name DamageName,
	float Momentum,
	vector HitLocation)
{
	local UGoldFixU227Base U227Module;

	if (SharedInfo != none && UGoldFix(SharedInfo.Owner) != none)
	{
		U227Module = UGoldFix(SharedInfo.Owner).U227Module;
		if (U227Module != none)
		{
			U227Module.Projectile_HurtRadiusProj(Proj, DamageAmount, DamageRadius, DamageName, Momentum, HitLocation);
			return;
		}
	}
	Proj.HurtRadius(DamageAmount, DamageRadius, DamageName, Momentum, HitLocation);
}

defaultproperties
{
				Version="UGoldFix version 7 release candidate 3 [2015-08-17]"
				VersionNumber=7
				VersionRevision=-1
				bEnableGameFix=True
				bEnableMapFix=True
				bAdjust_bStatic_Actors=True
				bAdjustBigRocks=True
				bAdjustBoulders=True
				bAdjustBrushes=True
				bAdjustCrushingMovers=True
				bAdjustDispersionFireRate=True
				bAdjustDistanceLightnings=True
				bAdjustEffects=True
				bAdjustEnhancedSightCheck=True
				bAdjustGLGrenades=True
				bAdjustIrrelevantMovers=True
				bAdjustMercenaries=True
				bAdjustMoverGroups=True
				bAdjustMoverMovements=True
				bAdjustPlayersPainTime=True
				bAdjustSpaceMarineCarcasses=True
				bAdjustSpaceMarineCARDamage=True
				bAdjustSpaceMarineName=True
				bAdjustSpaceMarines=True
				bAdjustSpawnedPlayerState=True
				bAdjustSpecialEvents=True
				bAdjustTriggerLights=True
				bAdjustViewingPlayers=True
				bCheckActorPackages=True
				bDisableMoversGoodCollision=True
				bExtraPlayerStartLookup=True
				bReplaceBlastDecals=True
				bReplaceExitUPakTeleporters=True
				bReplaceExplosiveBullets=True
				bReplaceFlashLightBeams=True
				bReplaceGrenades=True
				bReplaceOctagons=True
				bReplaceRockets=True
				bReplaceSpinnerProjectiles=True
				bReplaceUPakRockets=True
				bReplaceWarlordRockets=True
				bCoverHolesInLevelsGeometry=True
				bEnableDecorativeMapChanges=True
				bEnableLogicMapChanges=True
				bImproveNPCBehavior=True
				bModifyNetRelevance=True
				bPreventFastRetriggering=True
				bEnableMoverProtection=True
				bEnableExtraProtection=True
				bEnableStationaryEquipment=True
				StationaryEquipment="AsbestosSuit JumpBoots SCUBAGear SearchLight ToxinSuit UPakScubaGear"
				bDiscardItemsOnGameEnd_Unreal=True
				bDiscardItemsOnGameEnd_RTNP=True
}
